
#include "dlgwin32.h"
#include "debug.h"
#include "resource.h"

#include <process.h>
#include <assert.h>


///////////////////////////////////////////////////////////////////////////////


int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow ) {

 	::DialogBox(hInstance, MAKEINTRESOURCE(IDD_MAINWND), NULL, DialogProc );

	return 0;
}


///////////////////////////////////////////////////////////////////////////////


BOOL CALLBACK DialogProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam) {

	switch(uMsg) {
	HANDLE_DLG_MSG (hwnd, WM_INITDIALOG, OnInitDialog);
	HANDLE_DLG_MSG (hwnd, WM_COMMAND, OnCommand);
	}

	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////


BOOL OnInitDialog(HWND hwnd, HWND hWndFocus, LPARAM lParam) {

	SetDlgItemText ( hwnd, IDC_PATH, TEXT("C:\\foo.dat") );
	
	return TRUE;

}


///////////////////////////////////////////////////////////////////////////////


void OnCommand(HWND hwnd, int nID, HWND hWndCtl, UINT codeNotify) {

	switch(nID) {
	case IDC_READ: {

		TCHAR szPathFrom [ MAX_PATH ];
		Edit_GetText(GetDlgItem(hwnd, IDC_PATH), szPathFrom, sizeof (szPathFrom) );

		TCHAR szPathTo [MAX_PATH];
		if (!GetTempFileName ( TEXT("C:\\temp"), TEXT("adf"), 0, szPathTo ) ) {
			::MessageBox ( hwnd, TEXT("GetTempFileName failed."), TEXT("Error"), MB_OK | MB_ICONERROR );
			break;
		}

		SetDlgItemText ( hwnd, IDC_TO_FILE, szPathTo );

		BeginAsyncRead ( hwnd, szPathFrom, szPathTo );

		break;
	}
	case IDCANCEL:
		EndDialog(hwnd, 0);
		break;

	}
}


///////////////////////////////////////////////////////////////////////////////


VOID BeginAsyncRead ( HWND hwnd, LPTSTR pszPathFrom, LPTSTR pszPathTo ) {

	//
	// Open the file (FROM)
	//
	
	HANDLE hFile = CreateFile ( 
		pszPathFrom, 
		GENERIC_READ,
		NULL,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		NULL);

	if ( INVALID_HANDLE_VALUE == hFile ) {
		::MessageBox ( hwnd, TEXT("CreateFile failed - hFile."), TEXT("Error"), MB_OK | MB_ICONERROR);
		return;
	}

	//
	// Open the file (TO)
	//

	HANDLE hSaveFile = CreateFile (
		pszPathTo,
		GENERIC_WRITE,
		NULL,
		NULL,
		OPEN_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,
		NULL);

	if ( INVALID_HANDLE_VALUE == hSaveFile ) {
		::MessageBox ( hwnd, TEXT("CreateFile failed - hSaveFile."), TEXT("Error"), MB_OK | MB_ICONERROR);
		CloseHandle ( hFile );
		return;
	}

	
	//
	// Create an event object
	//

#ifdef USEEVENT

	HANDLE hEvent = CreateEvent (
		NULL, // Security Attributes
		TRUE, // Manual Reset is TRUE
		FALSE, // Initial State.
		NULL); // Name

	if ( !hEvent ) {
		::MessageBox ( hwnd, TEXT("Failed to create an event."), TEXT("Error"), MB_OK | MB_ICONERROR );
		CloseHandle ( hFile );
		CloseHandle ( hSaveFile );
		return;
	}

#endif

	//
	// Create a CONTEXT
	//
	
	PREAD_CONTEXT pContext = (PREAD_CONTEXT) malloc ( sizeof (READ_CONTEXT) );
	DWORD cbFileSize = GetFileSize ( hFile, NULL );
	
	if ( !pContext || INVALID_FILE_SIZE == cbFileSize ) {
		::MessageBox ( hwnd, TEXT("Failed to allocate the memory or get the file size."), TEXT("Error"), MB_OK | MB_ICONERROR);
		CloseHandle ( hFile );
		CloseHandle ( hSaveFile );
#ifdef USEEVENT
		CloseHandle ( hEvent );
#endif
		return;
	}
	ZeroMemory ( pContext , sizeof (READ_CONTEXT) );

	pContext->hFile = hFile;
	pContext->cbTotalBytesRead = 0;
	pContext->cbFileSize = cbFileSize;
	pContext->ol.Offset = 0;
	pContext->hSaveFile = hSaveFile;
#ifdef USEEVENT
	pContext->ol.hEvent = hEvent;
#endif

	
	//
	// Read file
	//

	BOOL bRet = ReadFile (
		hFile,
		pContext->pBuffer,
		BUFFER_SIZE,
		NULL,
		&pContext->ol );
		
	DWORD dwGle = GetLastError ();

	if ( bRet || (!bRet && ERROR_IO_PENDING == dwGle) ) {

		if (bRet) { // Just Tracing
			DebugPrint ( TEXT("ReadFile completed immediately.\n") );
		}

		// This is still in the main thread. The main thread is also the UI thread, which is
		// responsible for responding user's actions. So avoiding hanging up the ui I'll begin 
		// a new thread and read the data there. Basically if this was a kind of worker thread, 
		// a new thread isn't required.
		pContext->hThread = (HANDLE) _beginthreadex(
			NULL, 
			NULL, 
			WorkerThreadFunc, 
			(void*) pContext, 
			NULL, 
			&pContext->uThreadId);

		if ( !pContext->hThread ) {
			::MessageBox (hwnd, TEXT("Failed to begin a new thread."), TEXT("Error"), MB_OK | MB_ICONERROR );
			goto cleanup;
		}

		// OK
		return;
	}
	else {

		TCHAR szErrorMessage [256];
		wsprintf ( szErrorMessage, TEXT("ReadFile failed (%u)"), dwGle );
		::MessageBox ( hwnd, szErrorMessage, TEXT("Error"), MB_OK | MB_ICONERROR );
		
	}

cleanup:

	// Error
	CleanupContext ( pContext );

}


///////////////////////////////////////////////////////////////////////////////


VOID CleanupContext ( PREAD_CONTEXT pContext ) {

	if ( !pContext ) {
		return;
	}

	if ( pContext->hFile ) {
		CloseHandle ( pContext->hFile );
		pContext->hFile = NULL;
	}
	
	if ( pContext->hSaveFile ) {
		CloseHandle ( pContext->hSaveFile );
		pContext->hSaveFile = NULL;
	}

	if ( pContext->ol.hEvent ) {
		CloseHandle ( pContext->ol.hEvent );
		pContext->ol.hEvent = NULL;
	}

	free ( pContext );

}


///////////////////////////////////////////////////////////////////////////////


unsigned int __stdcall WorkerThreadFunc (PVOID pv) {

	DebugPrint ( TEXT("WorkerThreadFunc %08x Start\n"), GetCurrentThreadId() );
	
	PREAD_CONTEXT pContext = (PREAD_CONTEXT) pv;
	LPOVERLAPPED pol = NULL;
	DWORD cbNumberOfBytesTransferred = 0;
	BOOL bRet;
	
	while (1) {
		
		bRet = GetOverlappedResult ( pContext->hFile, &pContext->ol, &cbNumberOfBytesTransferred, TRUE );

		if ( !bRet ) {
			DebugPrint ( TEXT("GetOverlappedResult returned %u\n"), GetLastError() );
			CleanupContext ( pContext );			
			break;
		}

		DebugPrint ( TEXT("* GetOverlappedResult returned\nTransferred: %u\n"), cbNumberOfBytesTransferred );

		//
		// Update the context
		//

		pContext->cbTotalBytesRead += cbNumberOfBytesTransferred;
		pContext->ol.Offset += cbNumberOfBytesTransferred;
		
		//
		// Use the buffer
		//
		DWORD cbNumberOfBytesWritten = 0;
		BOOL bWrite;
		bWrite = WriteFile ( 
			pContext->hSaveFile, 
			pContext->pBuffer,
			cbNumberOfBytesTransferred,
			&cbNumberOfBytesWritten,
			NULL );

		if ( !bWrite ) {
			DebugPrint ( TEXT("WriteFile failed %u\n"), GetLastError() );
			CleanupContext ( pContext );
			break;
		}
		
		// DebugHexDump ( cbNumberOfBytesTransferred, (PBYTE) pContext->pBuffer );
		ZeroMemory ( pContext->pBuffer, BUFFER_SIZE );

		//
		// Need to read again?
		//

		if ( pContext->cbFileSize == pContext->cbTotalBytesRead ) {

			//
			// Read the file completely.
			//
			
			// Clean up the context
			//

			CleanupContext ( pContext );

			DebugPrint ( TEXT("OK!!\n") );

			// Exit worker thread.
			break;
		}
		
		//
		// Call ReadFile again.
		//

#ifdef USEEVENT

		ResetEvent ( pContext->ol.hEvent );

#endif

		bRet = ReadFile (
			pContext->hFile,
			pContext->pBuffer,
			min (BUFFER_SIZE, (pContext->cbFileSize - pContext->cbTotalBytesRead)),
			NULL,
			&pContext->ol );

		if ( !bRet ) {

			DWORD dwGle = GetLastError ();

			if ( ERROR_IO_PENDING != dwGle ) {

				CleanupContext ( pContext );
				DebugPrint ( TEXT("ERROR!! %u\n"), dwGle );
				
				break;
			}
		}
		
	}

	DebugPrint ( TEXT("WorkerThreadFunc %08x End\n"), GetCurrentThreadId() );

	return 0;
}

